#include <string.h>

#include "log.h"
#include "variant.h"

#define log_kStart 0
#define log_kInitialising 1
#define log_kIdle 2
#define log_kReading 3
#define log_kWriting 4

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-but-set-variable"

static struct pt pt;
typedef PT_THREAD((*updateFuncPtr)(void));
static updateFuncPtr updateFunction = NULL;

static nv_blob_t *readBlob;
static nv_blob_t *writeBlob;

uint16_t readAddress;
uint16_t writeAddress;

int8_t log_update(void) {
  if (updateFunction) {
    if (PT_SCHEDULE(updateFunction())) {
      return LOG_BUSY;
    }
  }
  return LOG_OK;
}

PT_THREAD(log_update_write(void)) {
  PT_BEGIN(&pt);
  // write this record
  PT_WAIT_THREAD(&pt, nv_write(writeAddress, writeBlob));
  // move to next record
  if (writeAddress) {
    writeAddress--;
  } else {
    writeAddress = nv_getSize() - 1;
  }
  // mark next record as free by erasing it
  PT_WAIT_THREAD(&pt, nv_erase(writeAddress));

  updateFunction = NULL;
  PT_END(&pt);
}

int8_t log_write(nv_blob_t *blob) {
  if (NULL == updateFunction) {
    writeBlob = blob;
    PT_INIT(&pt);
    updateFunction = log_update_write;
    log_update_write();
    return LOG_OK;
  }
  return LOG_BUSY;
}

PT_THREAD(log_update_init(void)) {
  char ptStatus;
  PT_BEGIN(&pt);
  nv_init();
  readAddress = nv_getSize();

  // look for empty record
  do {
    readAddress--;
    PT_WAIT_THREAD(&pt, ptStatus = nv_isFree(readAddress));
    if (ptStatus == PT_ENDED) {
      break;
    }
  } while (readAddress);

  // if no empty record found, create one
  if (ptStatus == PT_EXITED) {
    readAddress = nv_getSize() - 1;
    PT_WAIT_THREAD(&pt, nv_erase(readAddress));
  }

  writeAddress = readAddress; // rd and wr addr point to first empty record

  updateFunction = NULL;

  PT_END(&pt);
}

int8_t log_init(void) {
  PT_INIT(&pt);
  updateFunction = log_update_init;
  return LOG_OK;
}

int8_t log_reset(void) {
  if (NULL == updateFunction) {
    readAddress = writeAddress + 2;
    if (readAddress >= nv_getSize()) {
      readAddress = 0; // rap if passed end of memory
    }
    return LOG_OK;
  }
  return LOG_BUSY;
}

PT_THREAD(log_update_read(void)) {
  PT_BEGIN(&pt);
  // advance to previous blob (last in first out order)
  if (readAddress >= nv_getSize()) {
    readAddress = 0; // rap if end of eeprom
  }

  PT_WAIT_THREAD(&pt, nv_read(readAddress, readBlob));
  readAddress++;

  updateFunction = NULL;
  PT_END(&pt);
}

int8_t log_read(nv_blob_t *data) {
  if (NULL == updateFunction) {
    readBlob = data;
    PT_INIT(&pt);
    updateFunction = log_update_read;
    log_update_read();
    return LOG_OK;
  }
  return LOG_BUSY;
}

#pragma GCC diagnostic pop
