#include "Arduino.h" // for digitalWrite
#include <avr/interrupt.h>
#include <avr/io.h>
#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include <util/twi.h>

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif

#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#include "pins_arduino.h"
#include "twi.h"

#define TWI_MRX 1
#define TWI_MTX 2
#define TWI_SRX 3
#define TWI_STX 4
#define TWI_STOPPING 5

static volatile int8_t twi_state;
static uint8_t twi_slarw;

static uint8_t *twi_masterBuffer;
static volatile uint8_t twi_masterBufferIndex;
static uint8_t twi_masterBufferLength;
static volatile int8_t twi_error;

void twi_init(void) {
  twi_state = TWI_OK; // initialize state
  cbi(TWSR, TWPS0);   // initialize twi prescaler and bit rate
  cbi(TWSR, TWPS1);
  TWBR = ((F_CPU / TWI_FREQ) - 16) / 2;
  TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA); // enable twi, acks, and interrupt
}

int8_t twi_getStatus(void){
  if (TWI_STOPPING == twi_state) {
    if (0 == (TWCR & _BV(TWSTO))) {
      twi_state = TWI_OK;
      return twi_error;
    }
  }
  return TWI_BUSY;
}

void twi_read(uint8_t deviceAddress, uint8_t *data, uint8_t length) {
  if (TWI_OK == twi_state) {
    twi_masterBuffer = data;
    twi_masterBufferLength = length - 1;
    twi_state = TWI_MRX;
    twi_error = TWI_OK;        // reset error state (0xFF.. no error occured)
    twi_masterBufferIndex = 0; // initialize buffer iteration vars
    twi_slarw = TW_READ;       // build sla+w, slave device address + w bit
    twi_slarw |= deviceAddress << 1;
    // send start condition
    TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT) | _BV(TWSTA);
  }
}

void twi_write(uint8_t deviceAddress, uint8_t *data, uint8_t length) {
  if (TWI_OK == twi_state) {
    twi_masterBuffer = data;
    twi_masterBufferLength = length;
    twi_state = TWI_MTX;
    twi_error = TWI_OK;
    twi_masterBufferIndex = 0; // initialize buffer iteration vars
    twi_slarw = TW_WRITE;      // build sla+w, slave device address + w bit
    twi_slarw |= deviceAddress << 1;
    // send start condition
    TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT) | _BV(TWSTA);
  }
}

/*
 * Function twi_reply
 * Desc     sends byte or readys receive line
 * Input    ack: byte indicating to ack or to nack
 * Output   none
 */
void twi_reply(uint8_t ack) {
  // transmit master read ready signal, with or without ack
  if (ack) {
    TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWINT) | _BV(TWEA);
  } else {
    TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWINT);
  }
}

void twi_stop(void) {
  // send stop condition
  TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT) | _BV(TWSTO);
  twi_state = TWI_STOPPING; // update twi state
}

void twi_releaseBus(void) {
  TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT); // release bus
  twi_state = TWI_OK;                                    // update twi state
}

SIGNAL(TWI_vect) {
  switch (TW_STATUS) {

  // All Master
  case TW_START:     // sent start condition
  case TW_REP_START: // sent repeated start condition
    // copy device address and r/w bit to output register and ack
    TWDR = twi_slarw;
    twi_reply(1);
    break;

  // Master Transmitter
  case TW_MT_SLA_ACK:  // slave receiver acked address
  case TW_MT_DATA_ACK: // slave receiver acked data
    // if there is data to send, send it, otherwise stop
    if (twi_masterBufferIndex < twi_masterBufferLength) {
      // copy data to output register and ack
      TWDR = twi_masterBuffer[twi_masterBufferIndex++];
      twi_reply(1);
    } else {
      twi_stop();
    }
    break;
  case TW_MT_SLA_NACK:     // address sent, nack received
    twi_error = TWI_ERROR; // TW_MT_SLA_NACK;
    twi_stop();
    break;
  case TW_MT_DATA_NACK:    // data sent, nack received
    twi_error = TWI_ERROR; // TW_MT_DATA_NACK;
    twi_stop();
    break;
  case TW_MT_ARB_LOST:     // lost bus arbitration
    twi_error = TWI_ERROR; // TW_MT_ARB_LOST;
    twi_releaseBus();
    break;

  // Master Receiver
  case TW_MR_DATA_ACK: // data received, ack sent
    // put byte into buffer
    twi_masterBuffer[twi_masterBufferIndex++] = TWDR;
  case TW_MR_SLA_ACK: // address sent, ack received
    // ack if more bytes are expected, otherwise nack
    if (twi_masterBufferIndex < twi_masterBufferLength) {
      twi_reply(1);
    } else {
      twi_reply(0);
    }
    break;
  case TW_MR_DATA_NACK: // data received, nack sent
    // put final byte into buffer
    twi_masterBuffer[twi_masterBufferIndex++] = TWDR;

  case TW_MR_SLA_NACK: // address sent, nack received
    twi_stop();
    break;

  // All
  case TW_NO_INFO: // no state information
    break;
  case TW_BUS_ERROR:       // bus error, illegal stop/start
    twi_error = TWI_ERROR; // TW_BUS_ERROR;
    twi_stop();
    break;
  }
}
