#include <string.h>

#include "bus.h"
#include "frame.h"
#include "hal.h"

void bus_init(void) {
  // initialise bus status
  bus_status = bus_kRXReceiving;
  // initialise hardware - pin change flag, uart and timer
  hal_init();
}

uint8_t bus_read(uint8_t *buffer) {
  // poll receiver
  uint8_t status = hal_status();
  if (hal_rxComplete(status)) {
    uint8_t rxData = hal_rx();
    if (hal_rxError(status)) {
      rxData = 0x80;
    }
    if (bus_kTXSending == bus_status) {
      // transmitting
      if (rxData == *bus_bufferPtr++) {
        if (*bus_bufferPtr) {
          // send next byte
          hal_tx(*bus_bufferPtr);
          hal_txClear(status);
        } else {
          // transmit comlpete
          bus_status = bus_kTXCompleted;
        }
      } else {
        // abort transmit
        bus_status = bus_kTXAborted;
      }
    }
    if (bus_kRXReceiving == bus_status) {
      // receiving - only expect 7 bit data
      if ((bus_bufferPtr >= (bus_buffer + bus_kBufferSize)) ||
          (0x80 & rxData)) {
        // abort receive
        bus_status = bus_kRXAborted;
        bus_bufferPtr = bus_buffer;
      } else {
        // reset buffer if start of frame token received (':')
        if (frame_kStartOfFrame == rxData) {
          bus_bufferPtr = bus_buffer;
        }
        *bus_bufferPtr++ = rxData;
        // copy buffer and return it if end of frame token received ('\n')
        if (frame_kEndOfFrame == rxData) {
          *bus_bufferPtr = 0;
          if (frame_deframe(buffer, bus_buffer)) {
            bus_status = bus_kRXCompleted;
          } else {
            bus_status = bus_kRXAborted;
          }
          bus_bufferPtr = bus_buffer;
        }
      }
    }
  }
  // in process of sending
  if (bus_status == bus_kTXSending) {
    // error if transmitted byte not yet received back
    if (hal_txComplete(status)) {
      bus_status = bus_kTXAborted;
    }
  }
  // update bus inactivity timer
  if (hal_getInactiveTimer() > bus_holdoff) {
    // waiting to start transmision
    if (bus_status == bus_kTXWaiting) {
      // send first byte and clear tx comlete flag
      hal_tx(*bus_buffer);
      bus_status = bus_kTXSending;
    }
  }
  // returning to receive mode
  if ((bus_kTXCompleted == bus_status) || (bus_kTXAborted == bus_status)) {
    bus_bufferPtr = bus_buffer;
  }
  // if not transmitting send status message and return to receiving mode
  status = bus_status;
  if ((status != bus_kTXSending) && (status != bus_kTXWaiting)) {
    bus_status = bus_kRXReceiving;
  }
  return status;
}

uint8_t bus_write(uint8_t *buffer, uint8_t priority) {
  if (bus_kRXReceiving == bus_status) {
    bus_bufferPtr = bus_buffer;
    frame_enframe(bus_buffer, buffer);
    bus_holdoff = hal_random(priority);
    bus_status = bus_kTXWaiting;
    return bus_status;
  } else {
    return bus_kTXAborted;
  }
}
