#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-but-set-variable"

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/power.h>

#include "Arduino.h"

#include "ssr/ssr.h"

// start up time for ADC in 10 ms multiples (250 ms)
#define kStartUpTime (25)
#define kBusThreshold (20)

#define adc_kTemperatureMUX ((1 << REFS0) | 0x07)
#define adc_kBusLineAMUX ((1 << REFS0) | 0x00)
#define adc_kBusVoltageMUX ((1 << REFS0) | 0x06)
#define adc_kBatteryVoltageMUX ((1 << REFS0) | 0x01)

// 47K load resistor and Vishay NTCLE100E3104JB0
#define TEMP_TABLE_NTCLE100E3104JB0                                            \
  {                                                                            \
    1010, 1005, 999, 990, 979, 965, 947, 925, 899, 868, 831, 791, 745, 696,    \
        644, 590, 536, 483, 432, 384, 338, 298, 260, 227, 198, 172, 149, 130,  \
        113, 98, 86, 75, 65, 57, 50, 44, 39, 34, 30                            \
  }

// 47K load resistor and Murata NCP21WF104J03RA
#define TEMP_TABLE_NCP21WF104J03RA                                             \
  {                                                                            \
    1012, 1008, 1002, 993, 983, 969, 952, 930, 904, 873, 836, 794, 747, 696,   \
        642, 587, 531, 476, 423, 374, 328, 287, 250, 217, 188, 162, 140, 122,  \
        105, 92, 80, 70, 61, 54, 47, 42, 37, 33, 29                            \
  }

static const int16_t PROGMEM TEMP_LOOKUP_TABLE[] = TEMP_TABLE_NCP21WF104J03RA;

#define TEMP_LOOKUP_TABLE_SIZE (sizeof(TEMP_LOOKUP_TABLE) / sizeof(int16_t))

int16_t getTemp(int16_t adcRaw) {
  int16_t ref_lo, ref_hi, delta, temp;
  int8_t i;

  i = TEMP_LOOKUP_TABLE_SIZE;
  do {
    ref_lo = (int16_t)pgm_read_word(&TEMP_LOOKUP_TABLE[--i]);
  } while (((int16_t)adcRaw > ref_lo) && (i > 1));

  ref_hi = (int16_t)pgm_read_word(&TEMP_LOOKUP_TABLE[--i]);
  delta = ref_hi - ref_lo;
  temp = -400 + (i + 1) * 50 - (50 * ((int16_t)adcRaw - ref_lo)) / delta;

  return temp;
}

static volatile uint8_t adc_mux = 0;

static volatile int8_t adc_vcount = 0;
static volatile int16_t adc_vsum = 0;
static volatile int16_t adc_vavg = 0;
static volatile int16_t adc_vfilt = 0;

static volatile uint8_t adc_bcount = 0;
static volatile uint16_t adc_bsum = 0;
static volatile int16_t adc_blast = 0;
static volatile uint8_t adc_bsteps = 0;

static volatile int16_t pwm_filter = 0;

static volatile uint8_t adc_channel = 0;
static volatile uint16_t adc_value = 0;

static volatile uint16_t vbat_sum = 0;
static volatile uint8_t vbat_count = 0;
static volatile uint16_t vbus_sum = 0;
static volatile uint8_t vbus_count = 0;
static volatile uint8_t vbus_steps = 0;

static volatile uint16_t vbat_start = 0;
static volatile uint16_t vbat_avg = 0;

static volatile uint16_t vbus_avg = 0;

static volatile uint8_t zcd_cycleCounter = 0;

// settable parameters
static int16_t ledStatus = 0;
static volatile int16_t vbatRawTarget = 0;

static volatile int16_t temperature = 0;
static volatile uint16_t busVoltage = 0;
static volatile uint16_t busLineA = 0;

static volatile uint16_t zcd_period = 0;
static volatile uint16_t period = 0;
static volatile uint8_t cycle = 0;

static volatile uint8_t adc_bState = false;
static volatile uint16_t adc_bMaxCount = 0;
static volatile uint16_t adc_bMinCount = 0;
static volatile uint16_t vbus_maxCount = 0;
static volatile uint16_t vbus_minCount = 0;
static volatile uint16_t vbus_pwm = 0;

static volatile uint8_t vsys = 0;

static struct pt pt;

PT_THREAD(loopVar(struct pt *pt)) {

  PT_BEGIN(pt);

  PT_WAIT_WHILE(pt, cycle == zcd_cycleCounter);

  cycle = zcd_cycleCounter;

  // update battery voltage measurement
  cli();
  int16_t v = adc_vavg;
  sei();

  vbat_avg = v * 20;

  PT_YIELD(pt);

  cli();
  uint8_t count = vbus_count;
  vbus_count = 0;
  uint32_t sum = vbus_sum;
  vbus_sum = 0;
  sei();

  if (count) {
    vbus_avg = ((sum * 5L) / (uint32_t)count);
  }

  PT_YIELD(pt);

  vbus_pwm = 0;
  if ((vbus_maxCount)) {
    vbus_pwm = ((uint32_t)vbus_maxCount * 100L) /
               ((uint32_t)vbus_maxCount + (uint32_t)vbus_minCount);
  }

  PT_YIELD(pt);

  // update measurements
  cli(); // protect because values set during interrupt
  period = zcd_period;
  uint16_t value = adc_value;
  uint8_t channel = adc_channel;
  sei();

  if (adc_kTemperatureMUX == channel) {
    temperature = getTemp(value);
    adc_mux = adc_kBusVoltageMUX;
  } else if (adc_kBusVoltageMUX == channel) {
    busVoltage = value;
    adc_mux = adc_kTemperatureMUX;
  } else {
    adc_mux = adc_kTemperatureMUX;
  }

  PT_YIELD(pt);

  // update LED's every2 cycles or 40 ms
  if (cycle & 0x01) {
    uint8_t mask = led_update(ledStatus);
    mask &= 0x07;
    mask <<= 3;
    PORTD &= (~(0x07 << 3)) | mask;
    PORTD |= mask;
  }

  PT_END(pt);
}

void loopVariant(void) { loopVar(&pt); }

void initVariant(void) {

  // enable output pins for Charge Pump and LED's
  DDRD |= (1 << PORTD6) | (1 << PORTD5) | (1 << PORTD4) | (1 << PORTD3);

  // enable attenuation of 80 mV / bit (48 V sys)
  ATN_48V();
  vsys = 48;

  // ADC init
  adc_mux = adc_kBatteryVoltageMUX;
  ADMUX = adc_mux;
  ADCSRA = (1 << ADEN) | (1 << ADIE) | (1 << ADPS0) | (1 << ADPS2);
  DIDR0 = (1 << ADC0D) | (1 << ADC1D);
  ADCSRA |= (1 << ADSC);

  // ZCD interrupt init
  PCIFR |= (1 << PCIF0);
  PCICR |= (1 << PCIE0);
  PCMSK0 |= (1 << PCINT0);

  // set up SSR control delay on timer 1
  //  OCR1A = kNomLinePeriod / 4; // delay 1/4 cycle before updating SSR state
  OCR1B = kSampleDelay; // kNomLinePeriod / 2; // use nom period for half cycle
                        // interrupt

  ICR1 = kMaxLinePeriod; // timeout if no zero crossing detected in maximim line
                         // period

  TCCR1B = (1 << WGM13) | (1 << WGM12) |
           (1 << CS11);                  // CTC using ICR1 as top, 1/8 prescale
  TIMSK1 = (1 << ICIE1) | (1 << OCIE1B); // enable timer interrupts

  SSR_OFF(); // clear output on SSR

  sei();

  PT_INIT(&pt);

  // wait for a stable battery voltage measurement
  uint8_t i = kStartUpTime;
  uint8_t lastCycle = zcd_cycleCounter;
  while (i--) {
    while (zcd_cycleCounter == lastCycle) {
      loopVariant();
      getRequest(); // called to kick watchdog
    }
    lastCycle = zcd_cycleCounter;
  }

  // record battery voltage at start up
  vbat_start = vbat_avg;

  // set attenuation to suit system voltage
  if (vbat_start < (18000 / 4)) {
    // actual voltage is < 18 V -> set attenuation to 20 mV / bit (12 V sys)
    ATN_12V();
    vsys = 12;
  } else if (vbat_start < (36000 / 4)) {
    // actual voltage is < 36 V -> set attenuation to 40 mV / bit (24 V sys)
    ATN_24V();
    vsys = 24;
  }
  // otherwise leave attenuation set to 80 mV / bit (48 V sys)

  // wait for a stable measurement with new attenuation
  i = kStartUpTime;
  lastCycle = zcd_cycleCounter;
  while (i--) {
    while (zcd_cycleCounter == lastCycle) {
      loopVariant();
      getRequest(); // called to kick watchdog
    }
    lastCycle = zcd_cycleCounter;
  }
}

ISR(PCINT0_vect) {
  // only respond to the falling edge of ZCD input with blanking period (2 ms)
  if ((0 == (PINB & (1 << PORTB0))) && (TCNT1 > kBlankPeriod)) {
    uint16_t tmr = TCNT1;
    TCNT1 = 0;
    if (tmr > kMinLinePeriod) {
      // update line period measurement
      zcd_period = tmr;
    } else {
      zcd_period = 0;
    }
  }
}

ISR(TIMER1_COMPB_vect) {
  // update measurements
  vbat_count = adc_vcount;
  vbat_sum = adc_vsum;
  adc_vsum = 0;
  adc_vcount = 0;

  vbus_count = adc_bcount;
  vbus_sum = adc_bsum;
  adc_bsum = 0;
  adc_bcount = 0;

  vbus_steps = adc_bsteps;
  adc_bsteps = 0;

  vbus_maxCount = adc_bMaxCount;
  adc_bMaxCount = 0;

  vbus_minCount = adc_bMinCount;
  adc_bMinCount = 0;

  zcd_cycleCounter++;

  // calculate average vbat for the cycle
  if (vbat_count) {
    adc_vavg = vbat_sum / vbat_count;
  } else {
    adc_vavg = 0;
  }

  // filter battery voltage
  adc_vfilt += adc_vavg - adc_vfilt / 16;

  // update SSR state if good ac line detected
  if (zcd_period) {
    if (vbatRawTarget > 0) {
      // mode : normal voltage regulation
      if ((adc_vfilt > vbatRawTarget)) {
        SSR_ON();
      } else {
        SSR_OFF();
      }
    } else if (vbatRawTarget < 0) {
      // mode : test mode 1/8 duty cycle - 
      if(vbatRawTarget == kTestDutyCycle) {
          if ((zcd_cycleCounter & 0x07) == 0) {
            SSR_ON();
          } else {
            SSR_OFF();
          }
      } else if(vbatRawTarget == kFullDutyCycle){
        SSR_ON();
      } else {
        SSR_OFF();
      }
    } else {
      // mode : output forced off
      SSR_OFF();
    }
  } else {
    // mode : output forced off
    SSR_OFF();
  }
}

ISR(TIMER1_CAPT_vect) {
  // if we get here ac line out of range or not present
  SSR_OFF();
  zcd_period = 0;
}

ISR(ADC_vect) {
  // store adc result
  if (adc_kBatteryVoltageMUX == ADMUX) {
    adc_vsum += ADC;
    adc_vcount++;
    if (adc_mux == adc_kBatteryVoltageMUX) {
      adc_mux = adc_kBusLineAMUX;
    }
  } else if (adc_kBusLineAMUX == ADMUX) {
    int16_t adc_braw = ADC;
    int16_t adc_bdiff = adc_braw - adc_blast;
    if (adc_bdiff > kBusThreshold) {
      adc_bsteps++;
      adc_bState = true;
    }
    if (adc_bdiff < -kBusThreshold) {
      adc_bsteps++;
      adc_bState = false;
    }
    if (adc_bState == true) {
      adc_bMaxCount++;
    } else {
      adc_bMinCount++;
    }
    adc_blast = adc_braw;
    adc_bsum += adc_braw;
    adc_bcount++;
    if (adc_mux == adc_kBusLineAMUX) {
      adc_mux = adc_kBatteryVoltageMUX;
    }
  } else {
    adc_value = ADC;
    adc_channel = ADMUX;
    adc_mux = adc_kBatteryVoltageMUX;
  }

  // select next adc channel and start conversion
  ADMUX = adc_mux;
  ADCSRA |= (1 << ADSC);

  // toggle charge pump output
  PORTD ^= (1 << PORTD6);
}

int analogRead(uint8_t channel) {
  int returnValue = 0;

  if (channel == SSRSTATE) {
    returnValue = SSR_STATE ? 1 : 0;
  } else if (channel == VBATAVG) {
    returnValue = vbat_avg;
  } else if (channel == VBATSTART) {
    returnValue = vbat_start;
  } else if (channel == PERIOD) {
    returnValue = period;
  } else if (channel == CYCLE) {
    loopVariant();
    returnValue = cycle;
  } else if (channel == TEMPERATURE) {
    returnValue = temperature;
  } else if (channel == VBUS) {
    returnValue = busVoltage;
  } else if (channel == VBUSA) {
    returnValue = vbus_avg;
  } else if (channel == VBUSSTEPS) {
    returnValue = vbus_steps;
  } else if (channel == VBUSPWM) {
    returnValue = vbus_pwm;
  } else if (channel == VSYS) {
    returnValue = vsys;
  }

  return returnValue;
}

void analogWrite(uint8_t channel, int value) {
  if (LEDSTATUS == channel) {
    ledStatus = value;
  } else if (VBATTARGET == channel) {
    if (value > 0) {
      vbatRawTarget = ((int32_t)value * 16L) / 20L;
    } else {
      vbatRawTarget = value;
    }
  }
}

#pragma GCC diagnostic pop
